/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

//r_alias.c -- alias model rendering

#include "quakedef.h"

extern cvar_t r_drawflat, gl_overbright_models, gl_fullbrights, r_lerpmodels, r_lerpmove; // johnfitz

//up to 16 color translated skins
gltexture_t *playertextures[MAX_SCOREBOARD]; // johnfitz -- changed to an array of pointers
extern unsigned int r_meshindexbuffer;
extern unsigned int r_meshvertexbuffer;

#define NUMVERTEXNORMALS	162

float	r_avertexnormals[NUMVERTEXNORMALS][3] =
{
#include "anorms.h"
};

vec3_t	shadevector;

extern vec3_t	lightcolor; // johnfitz -- replaces "float shadelight" for lit support

// precalculated dot products for quantized angles
#define SHADEDOT_QUANT 16
float	r_avertexnormal_dots[SHADEDOT_QUANT][256] =
#include "anorm_dots.h"
	;

extern	vec3_t			lightspot;

qboolean	overbright; // johnfitz


/*
=================
R_SetupAliasFrame -- johnfitz -- rewritten to support lerping
=================
*/
void R_SetupAliasFrame (entity_t *e, aliashdr_t *paliashdr, int frame, aliasstate_t *state)
{
	int posenum, numposes;

	if ((frame >= paliashdr->numframes) || (frame < 0))
	{
		Con_DPrintf ("R_AliasSetupFrame: no such frame %d\n", frame);
		frame = 0;
	}

	posenum = paliashdr->frames[frame].firstpose;
	numposes = paliashdr->frames[frame].numposes;

	if (numposes > 1)
	{
		e->lerptime = paliashdr->frames[frame].interval;
		posenum += (int) (cl.time / e->lerptime) % numposes;
	}
	else
		e->lerptime = 0.1;

	if (e->lerpflags & LERP_RESETANIM) //kill any lerp in progress
	{
		e->lerpstart = 0;
		e->previouspose = posenum;
		e->currentpose = posenum;
		e->lerpflags -= LERP_RESETANIM;
	}
	else if (e->currentpose != posenum) // pose changed, start new lerp
	{
		if (e->lerpflags & LERP_RESETANIM2) //defer lerping one more time
		{
			e->lerpstart = 0;
			e->previouspose = posenum;
			e->currentpose = posenum;
			e->lerpflags -= LERP_RESETANIM2;
		}
		else
		{
			e->lerpstart = cl.time;
			e->previouspose = e->currentpose;
			e->currentpose = posenum;
		}
	}

	//set up values
	if (r_lerpmodels.value && !(e->model->flags &MOD_NOLERP && r_lerpmodels.value != 2))
	{
		if (e->lerpflags &LERP_FINISH && numposes == 1)
			state->blend = CLAMP (0, (cl.time - e->lerpstart) / (e->lerpfinish - e->lerpstart), 1);
		else
			state->blend = CLAMP (0, (cl.time - e->lerpstart) / e->lerptime, 1);

		state->pose1 = e->previouspose;
		state->pose2 = e->currentpose;
	}
	else //don't lerp
	{
		state->blend = 1;
		state->pose1 = posenum;
		state->pose2 = posenum;
	}
}

/*
=================
R_SetupEntityTransform -- johnfitz -- set up transform part of state
=================
*/
void R_SetupEntityTransform (entity_t *e, aliasstate_t *state)
{
	float blend;
	vec3_t d;
	int i;

	// if LERP_RESETMOVE, kill any lerps in progress
	if (e->lerpflags & LERP_RESETMOVE)
	{
		e->movelerpstart = 0;
		VectorCopy (e->origin, e->previousorigin);
		VectorCopy (e->origin, e->currentorigin);
		VectorCopy (e->angles, e->previousangles);
		VectorCopy (e->angles, e->currentangles);
		e->lerpflags -= LERP_RESETMOVE;
	}
	else if (!VectorCompare (e->origin, e->currentorigin) || !VectorCompare (e->angles, e->currentangles)) // origin/angles changed, start new lerp
	{
		e->movelerpstart = cl.time;
		VectorCopy (e->currentorigin, e->previousorigin);
		VectorCopy (e->origin,  e->currentorigin);
		VectorCopy (e->currentangles, e->previousangles);
		VectorCopy (e->angles,  e->currentangles);
	}

	//set up values
	if (r_lerpmove.value && e != &cl.viewent && e->lerpflags & LERP_MOVESTEP)
	{
		if (e->lerpflags & LERP_FINISH)
			blend = CLAMP (0, (cl.time - e->movelerpstart) / (e->lerpfinish - e->movelerpstart), 1);
		else blend = CLAMP (0, (cl.time - e->movelerpstart) / 0.1, 1);

		//translation
		VectorSubtract (e->currentorigin, e->previousorigin, d);
		state->origin[0] = e->previousorigin[0] + d[0] * blend;
		state->origin[1] = e->previousorigin[1] + d[1] * blend;
		state->origin[2] = e->previousorigin[2] + d[2] * blend;

		//rotation
		VectorSubtract (e->currentangles, e->previousangles, d);

		for (i = 0; i < 3; i++)
		{
			if (d[i] > 180)  d[i] -= 360;
			if (d[i] < -180) d[i] += 360;
		}

		state->angles[0] = e->previousangles[0] + d[0] * blend;
		state->angles[1] = e->previousangles[1] + d[1] * blend;
		state->angles[2] = e->previousangles[2] + d[2] * blend;
	}
	else //don't lerp
	{
		VectorCopy (e->origin, state->origin);
		VectorCopy (e->angles, state->angles);
	}
}

/*
=================
R_SetupAliasLighting -- johnfitz -- broken out from R_DrawAliasModel and rewritten
=================
*/
void R_SetupAliasLighting (entity_t	*e)
{
	float add;

	R_LightPoint (e);
	VectorCopy (lightspot, e->aliasstate.lightspot);

	// minimum light value on gun (24)
	if (e == &cl.viewent)
	{
		add = 72.0f - (lightcolor[0] + lightcolor[1] + lightcolor[2]);

		if (add > 0.0f)
		{
			lightcolor[0] += add / 3.0f;
			lightcolor[1] += add / 3.0f;
			lightcolor[2] += add / 3.0f;
		}
	}

	// minimum light value on players (8)
	if (e->entnum >= 1 && e->entnum <= cl.maxclients)
	{
		add = 24.0f - (lightcolor[0] + lightcolor[1] + lightcolor[2]);

		if (add > 0.0f)
		{
			lightcolor[0] += add / 3.0f;
			lightcolor[1] += add / 3.0f;
			lightcolor[2] += add / 3.0f;
		}
	}

	// clamp lighting so it doesn't overbright as much (96)
	if (overbright)
	{
		add = 288.0f / (lightcolor[0] + lightcolor[1] + lightcolor[2]);

		if (add < 1.0f)
			VectorScale (lightcolor, add, lightcolor);
	}

	//hack up the brightness when fullbrights but no overbrights (256)
	if (gl_fullbrights.value && !gl_overbright_models.value)
	{
		if (e->model->flags & MOD_FBRIGHTHACK)
		{
			lightcolor[0] = 256.0f;
			lightcolor[1] = 256.0f;
			lightcolor[2] = 256.0f;
		}
	}

	e->aliasstate.shadedots = r_avertexnormal_dots[((int) (e->angles[1] * (SHADEDOT_QUANT / 360.0))) & (SHADEDOT_QUANT - 1)];
	VectorScale (lightcolor, 1.0f / 200.0f, lightcolor);
}


// johnfitz -- values for shadow matrix
#define SHADOW_SKEW_X -0.7 //skew along x axis. -0.7 to mimic glquake shadows
#define SHADOW_SKEW_Y 0 //skew along y axis. 0 to mimic glquake shadows
#define SHADOW_VSCALE 0 //0=completely flat
#define SHADOW_HEIGHT 0.1 //how far above the floor to render the shadow
// johnfitz

float shadowmatrix16[16] =
{
	1,				0,				0,				0,
	0,				1,				0,				0,
	SHADOW_SKEW_X,	SHADOW_SKEW_Y,	SHADOW_VSCALE,	0,
	0,				0,				SHADOW_HEIGHT,	1
};

glmatrix shadowmatrix;

void R_DrawAliasShadow (entity_t *ent, aliashdr_t *hdr, aliasstate_t *state, float *dstmesh, float *shadecolor)
{
	int v;
	trivertx_t *verts1, *verts2;
	float	blend, iblend;
	qboolean lerping;
	aliasmesh_t *aliasmesh;
	trivertx_t *trivert1;
	trivertx_t *trivert2;

	memcpy (shadowmatrix.m16, shadowmatrix16, sizeof (float) * 16);

	aliasmesh = (aliasmesh_t *) ((byte *) hdr + hdr->aliasmesh);

	if (state->pose1 != state->pose2)
	{
		lerping = true;

		verts1 = (trivertx_t *) ((byte *) hdr + hdr->vertexes + hdr->framevertexsize * state->pose1);
		verts2 = (trivertx_t *) ((byte *) hdr + hdr->vertexes + hdr->framevertexsize * state->pose2);

		blend = state->blend;
		iblend = 1.0f - blend;
	}
	else // poses the same means either 1. the entity has paused its animation, or 2. r_lerpmodels is disabled
	{
		lerping = false;

		verts1 = (trivertx_t *) ((byte *) hdr + hdr->vertexes + hdr->framevertexsize * state->pose1);
		verts2 = verts1;

		blend = iblend = 0; // avoid bogus compiler warning
	}

	if (lerping)
	{
		for (v = 0; v < hdr->numverts; v++, dstmesh += 4, aliasmesh++)
		{
			trivert1 = &verts1[aliasmesh->vertindex];
			trivert2 = &verts2[aliasmesh->vertindex];

			dstmesh[0] = trivert1->v[0] * iblend + trivert2->v[0] * blend;
			dstmesh[1] = trivert1->v[1] * iblend + trivert2->v[1] * blend;
			dstmesh[2] = trivert1->v[2] * iblend + trivert2->v[2] * blend;
			dstmesh[3] = shadecolor[0];
		}
	}
	else
	{
		for (v = 0; v < hdr->numverts; v++, dstmesh += 4, aliasmesh++)
		{
			trivert1 = &verts1[aliasmesh->vertindex];

			dstmesh[0] = trivert1->v[0];
			dstmesh[1] = trivert1->v[1];
			dstmesh[2] = trivert1->v[2];
			dstmesh[3] = shadecolor[0];
		}
	}

	if (gl_arb_vertex_buffer_object && r_meshindexbuffer)
		GL_SetIndices (r_meshindexbuffer, (void *) hdr->firstindex);
	else GL_SetIndices (0, ((byte *) hdr + hdr->indexes));

	GL_DrawIndexedPrimitive (GL_TRIANGLES, hdr->numindexes, hdr->numverts);
}


void R_DrawAliasShadows (entity_t **ents, int numents, void *meshbuffer)
{
	if (r_shadows.value > 0.01f)
	{
		int i;
		qboolean stateset = false;
		byte shadecolor[4] = {0, 0, 0, 128};
		extern int gl_stencilbits;

		for (i = 0; i < numents; i++)
		{
			entity_t *ent = ents[i];
			glmatrix eshadow;

			if (!(ent->model->flags & MOD_NOSHADOW))
			{
				aliasstate_t *state = &ent->aliasstate;
				aliashdr_t *hdr = Mod_Extradata (ent->model);
				float lheight = state->origin[2] - state->lightspot[2];

				if (!stateset)
				{
					float *mesh = (float *) meshbuffer;

					glDepthMask (GL_FALSE);
					glEnable (GL_BLEND);
					GL_TexEnv (GL_TEXTURE1, GL_TEXTURE_2D, GL_NONE);
					GL_TexEnv (GL_TEXTURE0, GL_TEXTURE_2D, GL_NONE);
					glColor4f (0, 0, 0, r_shadows.value);
					shadecolor[3] = BYTE_CLAMPF (r_shadows.value);

					if (gl_stencilbits)
					{
						glEnable (GL_STENCIL_TEST);
						glStencilFunc (GL_EQUAL, 1, 2);
						glStencilOp (GL_KEEP, GL_KEEP, GL_INCR);
					}

					GL_SetStreamSource (0, GLSTREAM_POSITION, 3, GL_FLOAT, sizeof (float) * 4, mesh);
					GL_SetStreamSource (0, GLSTREAM_COLOR, 4, GL_UNSIGNED_BYTE, sizeof (float) * 4, &mesh[3]);
					GL_SetStreamSource (0, GLSTREAM_TEXCOORD0, 0, GL_NONE, 0, NULL);
					GL_SetStreamSource (0, GLSTREAM_TEXCOORD1, 0, GL_NONE, 0, NULL);
					GL_SetStreamSource (0, GLSTREAM_TEXCOORD2, 0, GL_NONE, 0, NULL);

					stateset = true;
				}

				GL_LoadMatrix (&eshadow, &r_world_matrix);

				if (state->origin[0] || state->origin[1] || state->origin[2])
					GL_TranslateMatrix (&eshadow, state->origin[0], state->origin[1], state->origin[2]);

				GL_TranslateMatrix (&eshadow, 0, 0, -lheight);
				GL_MultiplyMatrix (&eshadow, &shadowmatrix, &eshadow);
				GL_TranslateMatrix (&eshadow, 0, 0, lheight);

				if (state->angles[1]) GL_RotateMatrix (&eshadow, state->angles[1], 0, 0, 1);

				GL_TranslateMatrix (&eshadow, hdr->scale_origin[0], hdr->scale_origin[1], hdr->scale_origin[2]);
				GL_ScaleMatrix (&eshadow, hdr->scale[0], hdr->scale[1], hdr->scale[2]);

				glLoadMatrixf (eshadow.m16);

				R_DrawAliasShadow (ent, hdr, state, (float *) meshbuffer, (float *) shadecolor);
			}
		}

		if (stateset)
		{
			if (gl_stencilbits)
				glDisable (GL_STENCIL_TEST);

			GL_TexEnv (GL_TEXTURE0, GL_TEXTURE_2D, GL_REPLACE);
			glDisable (GL_BLEND);
			glDepthMask (GL_TRUE);
			glColor4f (1, 1, 1, 1);

			glLoadMatrixf (r_world_matrix.m16);
		}
	}
}


void R_SetupAliasModel (entity_t *e)
{
	aliashdr_t	*paliashdr;
	int			anim;
	aliasstate_t	*state = &e->aliasstate;
	vec3_t mins, maxs;

	memset (state, 0, sizeof (aliasstate_t));

	// initially not visible
	e->visframe = -1;

	// setup pose/lerp data -- do it first so we don't miss updates due to culling
	paliashdr = (aliashdr_t *) Mod_Extradata (e->model);
	R_SetupAliasFrame (e, paliashdr, e->frame, state);
	R_SetupEntityTransform (e, state);

	// we get blinky with per-frame so use the whole model bbox instead
	VectorAdd (state->origin, e->model->mins, mins);
	VectorAdd (state->origin, e->model->maxs, maxs);

	// hack - fixup scale
	//if (!e->scale) e->scale = 16;

	if (e->scale != 16)
	{
		// scale the bbox
		VectorScale (mins, (float) e->scale / 16.0f, mins);
		VectorScale (maxs, (float) e->scale / 16.0f, maxs);
	}

	// cull it (the viewmodel is never culled)
	if (r_shadows.value > 0.01f)
	{
		if (e != &cl.viewent)
		{
			if (R_CullBox (mins, maxs))
				e->visframe = -1;
			else e->visframe = r_framecount;
		}
	}
	else
	{
		if (e != &cl.viewent)
			if (R_CullBox (mins, maxs)) return;

		// the ent is visible now
		e->visframe = r_framecount;
	}

	// set up lighting
	overbright = gl_overbright_models.value;
	rs_aliaspolys += paliashdr->numtris;
	R_SetupAliasLighting (e);

	// store out the lighting
	state->shadelight[0] = lightcolor[0];
	state->shadelight[1] = lightcolor[1];
	state->shadelight[2] = lightcolor[2];
	state->shadelight[3] = e->alpha;

	// set up textures
	anim = (int) (cl.time * 10) & 3;

	// dead bodies retain correct colour; swiped from QuakeWorld; this was a lot easier than i thought it would be...
	if (e->scoreboard && playertextures[e->scoreboard->playertexture])
		state->tx = playertextures[e->scoreboard->playertexture];
	else state->tx = paliashdr->gltextures[e->skinnum][anim];

	if (!gl_fullbrights.value)
		state->fb = NULL;
	else state->fb = paliashdr->fbtextures[e->skinnum][anim];
}


void *r_meshbuffer = NULL;

typedef struct r_aliasmesh_s
{
	float xyz[3];

	union
	{
		byte color[4];
		unsigned int rgba;
	};
} r_aliasmesh_t;

void R_DrawAliasModel (entity_t *ent, aliashdr_t *hdr, aliasstate_t *state, qboolean showtris)
{
	int v;
	trivertx_t *verts1, *verts2;
	float	blend, iblend;
	qboolean lerping;
	aliasmesh_t *mesh;
	trivertx_t *trivert1;
	trivertx_t *trivert2;

	r_aliasmesh_t *r_aliasmesh;
	float *r_aliasst;

	// go back to the world matrix
	GL_LoadMatrix (&ent->matrix, &r_world_matrix);

	if (state->origin[0] || state->origin[1] || state->origin[2])
		GL_TranslateMatrix (&ent->matrix, state->origin[0], state->origin[1], state->origin[2]);

	if (state->angles[1]) GL_RotateMatrix (&ent->matrix, state->angles[1], 0, 0, 1);
	if (state->angles[0]) GL_RotateMatrix (&ent->matrix, -state->angles[0], 0, 1, 0);
	if (state->angles[2]) GL_RotateMatrix (&ent->matrix, state->angles[2], 1, 0, 0);

	GL_TranslateMatrix (&ent->matrix, hdr->scale_origin[0], hdr->scale_origin[1], hdr->scale_origin[2]);
	GL_ScaleMatrix (&ent->matrix, hdr->scale[0], hdr->scale[1], hdr->scale[2]);

	// DP_ENT_SCALE
	if (ent->scale != 16)
		GL_ScaleMatrix (&ent->matrix, (float) ent->scale / 16.0f, (float) ent->scale / 16.0f, (float) ent->scale / 16.0f);

	glLoadMatrixf (ent->matrix.m16);

	r_aliasmesh = (r_aliasmesh_t *) r_meshbuffer;
	r_aliasst = (float *) (r_aliasmesh + hdr->numverts);
	mesh = (aliasmesh_t *) ((byte *) hdr + hdr->aliasmesh);

	if (state->pose1 != state->pose2)
	{
		lerping = true;

		verts1 = (trivertx_t *) ((byte *) hdr + hdr->vertexes + hdr->framevertexsize * state->pose1);
		verts2 = (trivertx_t *) ((byte *) hdr + hdr->vertexes + hdr->framevertexsize * state->pose2);

		blend = state->blend;
		iblend = 1.0f - blend;
	}
	else // poses the same means either 1. the entity has paused its animation, or 2. r_lerpmodels is disabled
	{
		lerping = false;

		verts1 = (trivertx_t *) ((byte *) hdr + hdr->vertexes + hdr->framevertexsize * state->pose1);
		verts2 = verts1;

		blend = iblend = 0; // avoid bogus compiler warning
	}

	GL_SetStreamSource (0, GLSTREAM_POSITION, 3, GL_FLOAT, sizeof (r_aliasmesh_t), r_aliasmesh->xyz);
	GL_SetStreamSource (0, GLSTREAM_COLOR, 4, GL_UNSIGNED_BYTE, sizeof (r_aliasmesh_t), r_aliasmesh->color);

	if (showtris)
	{
		GL_SetStreamSource (0, GLSTREAM_TEXCOORD0, 0, GL_NONE, 0, NULL);
		GL_SetStreamSource (0, GLSTREAM_TEXCOORD1, 0, GL_NONE, 0, NULL);
	}
	else
	{
		if (gl_arb_vertex_buffer_object && r_meshvertexbuffer)
		{
			GL_SetStreamSource (r_meshvertexbuffer, GLSTREAM_TEXCOORD0, 2, GL_FLOAT, 0, (void *) hdr->firstvertex);

			if (state->fb)
				GL_SetStreamSource (r_meshvertexbuffer, GLSTREAM_TEXCOORD1, 2, GL_FLOAT, 0, (void *) hdr->firstvertex);
			else GL_SetStreamSource (0, GLSTREAM_TEXCOORD1, 0, GL_NONE, 0, NULL);
		}
		else
		{
			GL_SetStreamSource (0, GLSTREAM_TEXCOORD0, 2, GL_FLOAT, 0, r_aliasst);

			if (state->fb)
				GL_SetStreamSource (0, GLSTREAM_TEXCOORD1, 2, GL_FLOAT, 0, r_aliasst);
			else GL_SetStreamSource (0, GLSTREAM_TEXCOORD1, 0, GL_NONE, 0, NULL);
		}
	}

	GL_SetStreamSource (0, GLSTREAM_TEXCOORD2, 0, GL_NONE, 0, NULL);

	if (lerping)
	{
		for (v = 0; v < hdr->numverts; v++, r_aliasmesh++, mesh++)
		{
			trivert1 = &verts1[mesh->vertindex];
			trivert2 = &verts2[mesh->vertindex];

			mesh->light = (state->shadedots[trivert1->lightnormalindex] * iblend + state->shadedots[trivert2->lightnormalindex] * blend);

			r_aliasmesh->xyz[0] = trivert1->v[0] * iblend + trivert2->v[0] * blend;
			r_aliasmesh->xyz[1] = trivert1->v[1] * iblend + trivert2->v[1] * blend;
			r_aliasmesh->xyz[2] = trivert1->v[2] * iblend + trivert2->v[2] * blend;
		}
	}
	else
	{
		for (v = 0; v < hdr->numverts; v++, r_aliasmesh++, mesh++)
		{
			trivert1 = &verts1[mesh->vertindex];

			mesh->light = state->shadedots[trivert1->lightnormalindex];

			r_aliasmesh->xyz[0] = trivert1->v[0];
			r_aliasmesh->xyz[1] = trivert1->v[1];
			r_aliasmesh->xyz[2] = trivert1->v[2];
		}
	}

	// reset these for lighting
	r_aliasmesh = (r_aliasmesh_t *) r_meshbuffer;
	r_aliasst = (float *) (r_aliasmesh + hdr->numverts);
	mesh = (aliasmesh_t *) ((byte *) hdr + hdr->aliasmesh);

	if (showtris)
	{
		for (v = 0; v < hdr->numverts; v++, r_aliasmesh++, mesh++)
		{
			r_aliasmesh->rgba = 0xffffffff;
		}
	}
	else if (gl_arb_vertex_buffer_object && r_meshvertexbuffer)
	{
		for (v = 0; v < hdr->numverts; v++, r_aliasmesh++, mesh++, r_aliasst += 2)
		{
			r_aliasmesh->color[0] = BYTE_CLAMPF (mesh->light * state->shadelight[0]);
			r_aliasmesh->color[1] = BYTE_CLAMPF (mesh->light * state->shadelight[1]);
			r_aliasmesh->color[2] = BYTE_CLAMPF (mesh->light * state->shadelight[2]);
			r_aliasmesh->color[3] = ent->alpha;
		}
	}
	else
	{
		for (v = 0; v < hdr->numverts; v++, r_aliasmesh++, mesh++, r_aliasst += 2)
		{
			r_aliasst[0] = mesh->st[0];
			r_aliasst[1] = mesh->st[1];

			r_aliasmesh->color[0] = BYTE_CLAMPF (mesh->light * state->shadelight[0]);
			r_aliasmesh->color[1] = BYTE_CLAMPF (mesh->light * state->shadelight[1]);
			r_aliasmesh->color[2] = BYTE_CLAMPF (mesh->light * state->shadelight[2]);
			r_aliasmesh->color[3] = ent->alpha;
		}
	}

	if (gl_arb_vertex_buffer_object && r_meshindexbuffer)
		GL_SetIndices (r_meshindexbuffer, (void *) hdr->firstindex);
	else GL_SetIndices (0, ((byte *) hdr + hdr->indexes));

	GL_DrawIndexedPrimitive (GL_TRIANGLES, hdr->numindexes, hdr->numverts);
}


void R_PadAliasTMU (GLenum tmu, gltexture_t *tx)
{
	float hscale = 1;
	float vscale = 1;

	GL_ActiveTexture (tmu);
	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();

	if (tx->flags & TEXPREF_PAD)
	{
		hscale = (float) tx->source_width / (float) tx->width;
		vscale = (float) tx->source_height / (float) tx->height;
	}

	// now scale for padding (it might not be padded in which case the scale factors are 1)
	glScalef (hscale, vscale, 1);

	glMatrixMode (GL_MODELVIEW);
}


void R_UnpadAliasTMU (GLenum tmu)
{
	// undo pad scaling
	GL_ActiveTexture (tmu);
	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();
	glMatrixMode (GL_MODELVIEW);
}


void R_DrawAliasBatchPass (entity_t **ents, int numents, qboolean showtris)
{
	int i;
	gltexture_t *lasttexture = NULL;
	gltexture_t *lastfullbright = NULL;
	qboolean padtmu0 = false;
	qboolean padtmu1 = false;

	if (!numents) return;

	if (gl_smoothmodels.value) glShadeModel (GL_SMOOTH);
	if (gl_affinemodels.value) glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);

	if (showtris)
		GL_TexEnv (GL_TEXTURE0, GL_TEXTURE_2D, GL_NONE);
	else if (gl_overbright_models.value)
		GL_TexEnv (GL_TEXTURE0, GL_TEXTURE_2D, GL_RGB_SCALE);
	else GL_TexEnv (GL_TEXTURE0, GL_TEXTURE_2D, GL_MODULATE);

	for (i = 0; i < numents; i++)
	{
		entity_t *ent = ents[i];
		aliasstate_t *state = &ent->aliasstate;
		aliashdr_t *hdr = Mod_Extradata (ent->model);

		// we need a separate test for culling here as r_shadows mode doesn't cull
		if (ent->visframe != r_framecount) continue;

		if (!showtris && ((state->tx != lasttexture) || (state->fb != lastfullbright)))
		{
			if (state->fb)
			{
				GL_TexEnv (GL_TEXTURE1, GL_TEXTURE_2D, GL_ADD);
				GL_BindTexture (GL_TEXTURE1, state->fb);

				// scale for padding (always because it may not need padding so it should be scaled anyway)
				R_PadAliasTMU (GL_TEXTURE1, state->fb);
				padtmu1 = true;
			}
			else GL_TexEnv (GL_TEXTURE1, GL_TEXTURE_2D, GL_NONE);

			GL_BindTexture (GL_TEXTURE0, state->tx);

			// scale for padding (always because it may not need padding so it should be scaled anyway)
			R_PadAliasTMU (GL_TEXTURE0, state->tx);
			padtmu0 = true;

			lasttexture = state->tx;
			lastfullbright = state->fb;
		}

		// OK, this works better in GL... go figure...
		R_DrawAliasModel (ent, hdr, state, showtris);
	}

	R_UnpadAliasTMU (GL_TEXTURE1);
	GL_TexEnv (GL_TEXTURE1, GL_TEXTURE_2D, GL_NONE);

	R_UnpadAliasTMU (GL_TEXTURE0);
	GL_TexEnv (GL_TEXTURE0, GL_TEXTURE_2D, GL_REPLACE);

	glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glShadeModel (GL_FLAT);
	glColor4f (1, 1, 1, 1);

	if (gl_arb_vertex_buffer_object)
	{
		// explicitly take down the VBO streams if they were used
		GL_SetStreamSource (0, GLSTREAM_TEXCOORD0, 0, GL_NONE, 0, NULL);
		GL_SetStreamSource (0, GLSTREAM_TEXCOORD1, 0, GL_NONE, 0, NULL);
	}

	// go back to the world matrix
	glLoadMatrixf (r_world_matrix.m16);
}


void R_DrawAliasBatches (entity_t **ents, int numents, void *meshbuffer)
{
	r_meshbuffer = meshbuffer;
	R_DrawAliasBatchPass (ents, numents, false);

	if (r_showtris.value)
	{
		R_ShowTrisBegin ();
		R_DrawAliasBatchPass (ents, numents, true);
		R_ShowTrisEnd ();
	}
}


int R_DrawAliasModels_Sort (entity_t **e1, entity_t **e2)
{
	// sort by texture (the same model can have different textures)
	return (int) (e1[0]->aliasstate.tx - e2[0]->aliasstate.tx);
}


void R_AddEntityToAlphaList (entity_t *ent);

void R_DrawAliasModels (void)
{
	int i;
	int numalias;

	// temp buffer for storing ents for sorting
	entity_t **ents = (entity_t **) scratchbuf;

	if (!r_drawentities.value) return;

	for (i = 0, numalias = 0; i < cl_numvisedicts; i++)
	{
		entity_t *ent = cl_visedicts[i];

		if (!ent->model) continue;
		if (ent->model->type != mod_alias) continue;

		if (ent->alpha < 1)
			ent->alpha = 255;
		else if (ent->alpha < 255)
		{
			R_AddEntityToAlphaList (ent);
			continue;
		}

		R_SetupAliasModel (ent);

		// if it was culled or otherwise removed from the renderer we don't bother with it
		// (unless it's got shadows enabled in which case we have to :( )
		if (ent->visframe != r_framecount && r_shadows.value < 0.01f) continue;

		// store this model
		ents[numalias++] = ent;
	}

	if (!numalias) return;

	// sort the models by texture so that we can batch the bastards
	qsort (ents, numalias, sizeof (entity_t *), (int (*) (const void *, const void *)) R_DrawAliasModels_Sort);

	R_DrawAliasBatches (ents, numalias, (void *) (ents + numalias));
	R_DrawAliasShadows (ents, numalias, (void *) (ents + numalias));
}


/*
=============
R_DrawViewModel -- johnfitz -- gutted

mh - now goes through same draw routine as other mdls
=============
*/
void R_DrawViewModel (void)
{
	entity_t *e = &cl.viewent;

	if (!r_drawviewmodel.value || !r_drawentities.value || chase_active.value || envmap) return;
	if (cl.items & IT_INVISIBILITY || cl.stats[STAT_HEALTH] <= 0) return;

	if (!e->model) return;
	if (e->model->type != mod_alias) return;

	// hack the depth range to prevent wall-pokeing
	glDepthRange (QGL_DEPTH_3D_BEGIN, QGL_DEPTH_VM_END);

	R_SetupAliasModel (e);
	e->visframe = r_framecount;
	R_DrawAliasBatches (&e, 1, scratchbuf);

	// restore the hacked depth range (and do any other takedown we might ever need)
	glDepthRange (QGL_DEPTH_3D_BEGIN, QGL_DEPTH_3D_END);
}



